@@ -30,6 +30,7 @@
 
 #include <urcu/arch.h>
 #include <urcu/futex.h>
+#include <urcu/system.h>
 
 /*
  * Using attribute "weak" for __urcu_compat_futex_lock and
@@ -53,11 +54,11 @@ pthread_cond_t __urcu_compat_futex_cond = PTHREAD_COND_INITIALIZER;
 int compat_futex_noasync(int32_t *uaddr, int op, int32_t val,
 	const struct timespec *timeout, int32_t *uaddr2, int32_t val3)
 {
-	int ret, gret = 0;
+	int ret = 0, lockret;
 
 	/*
 	 * Check if NULL. Don't let users expect that they are taken into
-	 * account. 
+	 * account.
 	 */
 	assert(!timeout);
 	assert(!uaddr2);
@@ -68,8 +69,12 @@ int compat_futex_noasync(int32_t *uaddr, int op, int32_t val,
 	 */
 	cmm_smp_mb();
 
-	ret = pthread_mutex_lock(&__urcu_compat_futex_lock);
-	assert(!ret);
+	lockret = pthread_mutex_lock(&__urcu_compat_futex_lock);
+	if (lockret) {
+		errno = lockret;
+		ret = -1;
+		goto end;
+	}
 	switch (op) {
 	case FUTEX_WAIT:
 		/*
@@ -77,7 +82,7 @@ int compat_futex_noasync(int32_t *uaddr, int op, int32_t val,
 		 * Comparing *uaddr content against val figures out which
 		 * thread has been awakened.
 		 */
-		while (*uaddr == val)
+		while (CMM_LOAD_SHARED(*uaddr) == val)
 			pthread_cond_wait(&__urcu_compat_futex_cond,
 				&__urcu_compat_futex_lock);
 		break;
@@ -90,11 +95,16 @@ int compat_futex_noasync(int32_t *uaddr, int op, int32_t val,
 		pthread_cond_broadcast(&__urcu_compat_futex_cond);
 		break;
 	default:
-		gret = -EINVAL;
+		errno = EINVAL;
+		ret = -1;
 	}
-	ret = pthread_mutex_unlock(&__urcu_compat_futex_lock);
-	assert(!ret);
-	return gret;
+	lockret = pthread_mutex_unlock(&__urcu_compat_futex_lock);
+	if (lockret) {
+		errno = lockret;
+		ret = -1;
+	}
+end:
+	return ret;
 }
 
 /*
@@ -106,9 +116,11 @@ int compat_futex_noasync(int32_t *uaddr, int op, int32_t val,
 int compat_futex_async(int32_t *uaddr, int op, int32_t val,
 	const struct timespec *timeout, int32_t *uaddr2, int32_t val3)
 {
+	int ret = 0;
+
 	/*
 	 * Check if NULL. Don't let users expect that they are taken into
-	 * account. 
+	 * account.
 	 */
 	assert(!timeout);
 	assert(!uaddr2);
@@ -121,13 +133,20 @@ int compat_futex_async(int32_t *uaddr, int op, int32_t val,
 
 	switch (op) {
 	case FUTEX_WAIT:
-		while (*uaddr == val)
-			poll(NULL, 0, 10);
+		while (CMM_LOAD_SHARED(*uaddr) == val) {
+			if (poll(NULL, 0, 10) < 0) {
+				ret = -1;
+				/* Keep poll errno. Caller handles EINTR. */
+				goto end;
+			}
+		}
 		break;
 	case FUTEX_WAKE:
 		break;
 	default:
-		return -EINVAL;
+		errno = EINVAL;
+		ret = -1;
 	}
-	return 0;
+end:
+	return ret;
 }
